#include <iostream>
#include <string>
#include <sstream>

#include "ndarray.h"

#include <vector>

#include <cmath>
#include <algorithm>
#include <iomanip>

using namespace std;

class Video {
        public:
                Video(istream&, int width, int height);
                int length() const; // in frames
                unsigned char& get(int frame, int row, int col, int channel);
        private:
                int frames, rows, cols;
                ndarray<unsigned char> pixels;
                ndarray<double> energy;
                
        public: void dump_bytes(ostream &);

        public: void carve_frame();

        double intensity(int,int,int);
};

Video::Video(istream& input, int width, int height)
: rows{height}, cols{width} {
        vector<unsigned char> all_bytes;
        int framesize = width*height*3; // bytes in a single frame
        char* buffer = new char[framesize];
        while(input.read(buffer, framesize)) {
                all_bytes.insert(all_bytes.end(), buffer, buffer + framesize);
        }
        delete[] buffer;
        clog << "Setting up: ";
        // determine how many frames we have
        frames = all_bytes.size() / (3*rows * cols);
        // then put the data into pixels
        ndarray<unsigned char> reshaped {frames, rows, cols, 3};
        int indx = 0; // index into 1d vector
        for (int f=0;f<frames;++f) {
        for (int r=0;r<rows;++r) {
        for (int c=0;c<cols;++c) {
        for (int chan=0;chan<3;++chan) {
                reshaped.get(f,r,c,chan) = all_bytes[indx];
                ++indx;
        }}}}
        pixels = reshaped;
        ndarray<double> temp {frames, cols};
        for (int r=0; r<rows; ++r) {
        int percentage = 100 * (r+1) / rows;
        if (r > 0) cerr << "\b\b\b\b";
        cerr << setw(3) << percentage << "%";
        if (r == rows-1) clog << "\n";

        for (int f=0; f<frames; ++f) {
        for (int c=0; c<cols; ++c) {
                double df = intensity(f,r,c) - intensity(f == 0 ? 1 : f-1, r,c);
                temp.get(f,c) = max(abs(df), temp.get(f,c));
        }}}
        energy = temp;

}
int Video::length() const {return frames;}
unsigned char & Video::get(int frame, int row, int col, int channel) {
        return pixels.get(frame, row, col, channel);
}

void Video::dump_bytes(ostream & out) {
        for (int f=0;f<length();++f) {
        //for (int r=0;r<rows;++r) {
        //for (int c=0;c<cols;++c) {
        //for (int chan=0;chan<3;++chan) {
        out.write((char *)&pixels.get(f), rows*cols*3);
        //      out.put(pixels.get(f,r,c,chan));
        }//}}}
}
void Video::carve_frame() {
        ndarray<double> opt {frames, cols};
        ndarray<int> prev {frames, cols};
        ndarray<int> to_delete {cols}; // mark frame to delete
        
        for (int f=0; f < frames; ++f)
                opt.get(f,0) = energy.get(f,0);
                
        for (int c=0; c+1<cols; ++c) {
        for (int f=0; f<frames; ++f) {
                opt.get(f,c+1) = opt.get(f,c);
                prev.get(f,c+1) = f;
                if (f-1 >= 0 && opt.get(f-1,c) < opt.get(f,c+1)) {
                        opt.get(f,c+1) = opt.get(f-1,c);
                        prev.get(f,c+1) = f-1;
                }
                if (f+1 < frames && opt.get(f+1,c) < opt.get(f,c+1)) {
                        opt.get(f,c+1) = opt.get(f+1,c);
                        prev.get(f,c+1) = f+1;
                }
                opt.get(f,c+1) += energy.get(f,c+1);
        }}

        // find endpoint of seam
        int f_min = 0;
        for (int f=1; f < frames; ++f) {
                if (opt.get(f,cols-1) < opt.get(f_min,cols-1))
                        f_min = f;
        }
        // trace through seam and mark pixels
        for (int c=cols-1; c >= 0; --c) {
                to_delete.get(c) = f_min;
                f_min = prev.get(f_min, c);
        }
        ndarray<unsigned char> new_pixels {frames-1, rows, cols, 3};
        ndarray<double> new_energy {frames-1, cols};
        for (int c=0; c<cols; ++c) {
                int f_new = 0; // frame index into replacement arrays
                for (int f_old = 0; f_old < frames; ++f_old) {
                        // f_old: frame index into existing arrays
                        if (f_old == to_delete.get(c)) continue;
                        for (int chan=0; chan<3; ++chan) {
                        for (int r=0; r<rows; ++r) {
                                new_pixels.get(f_new,r,c,chan) =
                                        pixels.get(f_old, r, c, chan);
                        }}
                        new_energy.get(f_new,c) =
                                energy.get(f_old,c);
                        ++f_new;
                }       
        }
        // maintain class invariant
        frames = frames-1;
        pixels = new_pixels;
        energy = new_energy;

}
double Video::intensity(int f, int r, int c) {
        double res {0.0};
        for (int chan=0; chan<3;++chan) res += get(f,r,c,chan);
        return res;
}

int main(int argc, const char * argv[]) {
        string expected {"Expected arguments: [width] [height] [fraction kept]\n"};
        if (argc != 4) {
                cout << expected;
                return 1;
        }

        istringstream width_string {argv[1]};
        istringstream height_string {argv[2]};
        istringstream fraction_string {argv[3]};
        double fraction; int width, height;
        if (
                !(fraction_string >> fraction) ||
                !(width_string >> width) ||
                !(height_string >> height)
        ) {
                cout << expected;
                cout << "I couldn't parse the arguments (int int double).\n";
                return 1;
        }
        if (fraction <= 0 || fraction > 1) {
        cout << "I require 0 < fraction <= 1.\n";
        return 1;
        }

        Video input_video {cin, width, height};

        int frames_to_remove = input_video.length() * (1.0 - fraction);
        for (int k=0; k < frames_to_remove; ++k) {
                if (k > 0) cerr << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b";
                cerr << "Frame: " << setw(4) << (k+1) << "/" << setw(4) << frames_to_remove;
                input_video.carve_frame();
        }
        clog << "\n";

        input_video.dump_bytes(cout);

        return 0;
}

